import * as superagent from 'superagent'
import * as firebase from "@firebase/testing";
import {WhereFilterOp} from "@firebase/firestore-types";
import {FirestoreClient} from "../FirestoreClient"
import {cursor} from "sisteransi";
import up = cursor.up;

const projectId = "ocf-integration-test"
const host = `http://localhost:5001/${projectId}/us-central1`
const crudFunctionName = "test_trigger_http_fs_v1 "

const collectionName = 'test'
let firestore
let fireStoreClient : FirestoreClient

describe('Integration Tests (offline using emulators)', () => {

    afterEach(async () => {
        await firebase.clearFirestoreData({projectId: projectId});
        firebase.apps().forEach(async app => await app.delete());
    });

    beforeEach(async () => {
        firestore = firebase.initializeTestApp({
            projectId: projectId
        }).firestore();
        fireStoreClient = new FirestoreClient(firestore)
    })

    it('create doc, update it and check version update', async () => {
        //given
        const entityForCreation = {someMandatoryAttrib: "someVal", someOptionalAttrib: "someOptionalVal"} as EntityForCreation

        //when
        let response = await addDocument(entityForCreation)

        //then
        expect(response).toBeDefined()
        let entity : Entity = response.body
        expect(entity.created).toBeDefined()
        expect(entity.lastModified).toBeDefined()
        expect(entity.version).toBe(1)

        //and the entity is updated
        entity.someMandatoryAttrib = "someOtherValue"
        delete entity.someOptionalAttrib

        response = await setDocument(entity.id, entity, ["someOptionalAttrib"])

        //then
        expect(response).toBeDefined()
        entity = response.body
        expect(entity.created).toBeDefined()
        expect(entity.lastModified).toBeDefined()
        expect(entity.version).toBe(2)
        expect(entity.someMandatoryAttrib).toBe("someOtherValue")
        expect(entity.someOptionalAttrib).toBeUndefined()

    })

    it('create doc, delete and checkIfExists', async () => {
        //given
        const entityForCreation = {someMandatoryAttrib: "someVal", someOptionalAttrib: "someOptionalVal"} as EntityForCreation

        //when
        let response = await addDocument(entityForCreation)

        //then
        expect(response).toBeDefined()
        let entity : Entity = response.body
        expect(await documentExists(entity.id)).toBeTruthy

        //and the entity is deleted
        response = await deleteDocument(entity.id)

        //then
        expect(await documentExists(entity.id)).toBeFalsy
    })


    it('bulk import of 1002 docs and execute query', async () => {
        //given
        const data = [] as EntityForCreation[]

        for (let i = 0; i < 1002; i++) {
            if (i == 1 || i == 5 || i == 10) {
                data.push({someMandatoryAttrib: "someFilteredVal", someOptionalAttrib: "someOptionalVal"} as EntityForCreation)
            } else {
                data.push({someMandatoryAttrib: "someVal", someOptionalAttrib: "someOptionalVal"} as EntityForCreation)
            }
        }

        //when
        await bulkSet(data)
        const response = await executeQuery("someMandatoryAttrib", "==", "someVal")

        //then
        expect(response).toBeDefined()
        expect(response.body.length).toBe(999)
    })

    it('create doc, than update within transaction', async () => {
        //given
        const entityForCreation = {someMandatoryAttrib: "someVal", someOptionalAttrib: "someOptionalVal"} as EntityForCreation
        let response = await addDocument(entityForCreation)
        let entity : Entity = response.body
        expect(entity.version).toBe(1)

        //when
        const result = await updateIdInTransaction(entity.id)

        //then
        response = await getDocument(entity.id)
        entity  = response.body
        expect(entity.version).toBe(2)
    })

    it('create docs and receive in bulk', async () => {
        //given
        const entityForCreation = {someMandatoryAttrib: "someVal", someOptionalAttrib: "someOptionalVal"} as EntityForCreation
        let response = await addDocument(entityForCreation)
        let entity : Entity = response.body
        expect(entity.version).toBe(1)

        const entityForCreation2 = {someMandatoryAttrib: "someVal", someOptionalAttrib: "someOptionalVal"} as EntityForCreation
        response = await addDocument(entityForCreation)
        let entity2 : Entity  = response.body
        expect(entity2.version).toBe(1)

        //when
        const result = await getDocumentsById([entity.id, entity2.id])

        //then
        let docs : Entity[]  = result.body
        expect(docs.length).toBe(2)
    })

    it('create doc, update and check', async () => {
        //given
        const entityForCreation = {someMandatoryAttrib: "someVal", someOptionalAttrib: "someOptionalVal"} as EntityForCreation

        //when
        let response = await addDocument(entityForCreation)

        //then
        expect(response).toBeDefined()
        let entity : Entity = response.body
        expect(await documentExists(entity.id)).toBeTruthy

        //and the entity is deleted
        response = await updateDocument(entity.id, {someOptionalAttrib: "someOtherOptionalVal"})
        expect(response.body.someOptionalAttrib).toBe("someOtherOptionalVal")
        expect(response.body.version).toBe(2)

    })

})


async function executeQuery(fieldPath: string, opStr: WhereFilterOp, value: any) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "executeQuery",
            args: [fieldPath, opStr, value]
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}
async function updateIdInTransaction(id: string) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            updateIdInTransaction: id
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}

async function getDocumentsById(ids: string[]) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "getDocumentsById",
            args: [ids]
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}

async function updateDocument(id: string, updateData: any) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "updateDocument",
            args: [id, updateData]
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}


async function documentExists(id: string) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "documentExists",
            args: [id]
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}

async function bulkSet(data: any[]) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "bulkSet",
            args: [data]
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}

async function deleteDocument(id: string) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "deleteDocument",
            args: [id]
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}

async function addDocument(data: any) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "addDocument",
            args: [data]
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}

async function setDocument(id: string, data: any, optionalFields : string[]  = []) {
    await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "setDocument",
            args: [
                id,
                data
            ],
            optionalFields: optionalFields
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
    return getDocument(id)
}

async function getDocument(id: string) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "getDocument",
            args: [
                id
            ]
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}


interface  EntityForCreation {
    someMandatoryAttrib : string,
    someOptionalAttrib?: string
}

interface  Entity extends EntityForCreation{
    created: Date,
    lastModified: Date,
    id : string
    version: number
}
